use std::fs::{File, OpenOptions};
use std::io::{Cursor, Read, Seek, SeekFrom, Write};
use std::path::Path;

use prost::Message;

pub use crate::proto::Entry;

pub type Error = std::io::Error;
pub type Result<T> = std::result::Result<T, Error>;

const X25: crc::Crc<u16> = crc::Crc::<u16>::new(&crc::CRC_16_IBM_SDLC);

// data is stored in append-only logs
// each data entry starts with a checksum (CRC16),
// followed by a one-byte length, and the
// actual entry encoded in protocol buffer
// upon opening a file, all the entries within it are
// read, any corrupted contents at the end truncated/overwritten
pub struct Store<T> {
    contents: T,
    entries: Vec<Entry>,
    start_ts: Option<u64>,
    end_ts: Option<u64>,
}

impl<T> Store<T> {
    pub fn num_entries(&self) -> usize {
        self.entries.len()
    }

    pub fn data(&self) -> Vec<Entry> {
        self.entries.clone()
    }

    pub fn start(&self) -> Option<u64> {
        self.start_ts
    }

    pub fn end(&self) -> Option<u64> {
        self.end_ts
    }
}

impl Store<Cursor<Vec<u8>>> {
    pub fn open_read_only<P: AsRef<Path>>(p: P) -> Result<Store<Cursor<Vec<u8>>>> {
        let data = std::fs::read(p)?;

        Ok(Store::new(Cursor::new(data)))
    }
}

impl Store<File> {
    pub fn open<P: AsRef<Path>>(p: P) -> Result<Store<File>> {
        let mut f = OpenOptions::new()
            .read(true)
            .write(true)
            .create(true)
            .open(p)?;
        let _ = f.seek(SeekFrom::Start(0))?;

        Ok(Store::new(f))
    }
}

impl<T: Read + Seek> Store<T> {
    pub fn new(mut t: T) -> Store<T> {
        let mut entries = Vec::new();
        let mut start_ts = None;
        let mut end_ts = None;

        // start of the next unread entry
        let mut cur_pos = 0;
        let mut data: Vec<u8> = Vec::new();
        loop {
            let mut crc16 = [0u8; 2];
            let mut len = [0u8; 1];
            if let Err(_) = t.read_exact(&mut crc16) {
                break;
            }
            if let Err(_) = t.read_exact(&mut len) {
                break;
            }

            data.resize(len[0] as usize, 0);
            if let Err(_) = t.read_exact(&mut data) {
                break;
            }

            // CRC check data
            let calculated = X25.checksum(&data);
            let crc = (crc16[0] as u16) | ((crc16[1] as u16) << 8);
            if calculated != crc {
                break;
            }

            // decode data into an entry
            if let Ok(entry) = Entry::decode(&data[..]) {
                cur_pos = t.stream_position().unwrap();

                if start_ts.is_none() {
                    start_ts = entry.timestamp.clone();
                }
                if entry.timestamp.is_some() {
                    end_ts = entry.timestamp.clone();
                }
                entries.push(entry);
            } else {
                break;
            }
        }

        let _ = t.seek(SeekFrom::Start(cur_pos)).unwrap();

        Store {
            contents: t,
            entries,
            start_ts,
            end_ts,
        }
    }

    pub fn lookup_range(&self, start: u64, end: Option<u64>) -> Vec<Entry> {
        if let Some(start_ts) = self.start_ts {
            if let Some(end_range) = end {
                if end_range < start_ts {
                    return Vec::new();
                }
            }
        }
        if let Some(end_ts) = self.end_ts {
            if start > end_ts {
                return Vec::new();
            }
        }
        let mut ret = Vec::new();
        for e in &self.entries {
            if let Some(ts) = e.timestamp {
                if ts >= start {
                    if let Some(end_range) = end {
                        if ts > end_range {
                            continue;
                        }
                    }
                    ret.push(e.clone());
                }
            }
        }

        ret
    }
}

impl<T: Write> Store<T> {
    pub fn append(&mut self, e: &Entry) -> Result<()> {
        let payload = e.encode_to_vec();
        let len = payload.len();
        let len_u8: [u8; 1] = if len < 256 {
            [len as u8]
        } else {
            return Err(Error::new(std::io::ErrorKind::Other, "payload too large"));
        };

        // CRC check data
        let calculated = X25.checksum(&payload);
        let crc: [u8; 2] = [(calculated & 0xff) as u8, ((calculated >> 8) & 0xff) as u8];

        self.contents.write(&crc)?;
        self.contents.write(&len_u8)?;
        self.contents.write(&payload)?;

        self.entries.push(e.clone());
        if self.start_ts.is_none() {
            self.start_ts = e.timestamp;
        }
        if let Some(ts) = e.timestamp {
            self.end_ts = Some(ts);
        }

        Ok(())
    }
}

#[cfg(test)]
mod tests {
    use super::{Entry, Store};

    #[test]
    fn append_read() {
        let temp_file = tempfile::NamedTempFile::new().unwrap();
        let entry = Entry {
            timestamp: Some(7),
            humidity: vec![0.95],
            temp: Some(32.2),
            fan: Some(false),
            humidifer: vec![false],
        };

        {
            let mut store = Store::open(temp_file.path()).unwrap();
            store.append(&entry).unwrap();

            let entries = store.lookup_range(0, None);
            assert_eq!(entries, vec![entry.clone()]);
        }

        {
            let store = Store::open_read_only(temp_file.path()).unwrap();
            let entries = store.lookup_range(0, None);
            assert_eq!(entries, vec![entry.clone()]);
        }
    }

    #[test]
    fn append2_read2() {
        let temp_file = tempfile::NamedTempFile::new().unwrap();
        let entry = Entry {
            timestamp: Some(7),
            humidity: vec![0.95],
            temp: Some(32.2),
            fan: Some(false),
            humidifer: vec![false],
        };

        let entry2 = Entry {
            timestamp: Some(16),
            humidity: vec![0.92],
            temp: Some(31.2),
            fan: Some(true),
            humidifer: vec![false],
        };

        {
            let mut store = Store::open(temp_file.path()).unwrap();
            store.append(&entry).unwrap();
            store.append(&entry2).unwrap();

            let entries = store.lookup_range(0, None);
            assert_eq!(entries, vec![entry.clone(), entry2.clone()]);
        }

        {
            let store = Store::open_read_only(temp_file.path()).unwrap();
            let entries = store.lookup_range(0, None);
            assert_eq!(entries, vec![entry.clone(), entry2.clone()]);
        }
    }

    #[test]
    fn append2_read2_interrupted() {
        let temp_file = tempfile::NamedTempFile::new().unwrap();
        let entry = Entry {
            timestamp: Some(7),
            humidity: vec![0.95],
            temp: Some(32.2),
            fan: Some(false),
            humidifer: vec![false],
        };

        let entry2 = Entry {
            timestamp: Some(16),
            humidity: vec![0.92],
            temp: Some(31.2),
            fan: Some(true),
            humidifer: vec![false],
        };

        {
            let mut store = Store::open(temp_file.path()).unwrap();
            store.append(&entry).unwrap();

            let entries = store.lookup_range(0, None);
            assert_eq!(entries, vec![entry.clone()]);
        }
        {
            let mut store = Store::open(temp_file.path()).unwrap();
            store.append(&entry2).unwrap();

            let entries = store.lookup_range(0, None);
            assert_eq!(entries, vec![entry.clone(), entry2.clone()]);
        }

        {
            let store = Store::open_read_only(temp_file.path()).unwrap();
            let entries = store.lookup_range(0, None);
            assert_eq!(entries, vec![entry.clone(), entry2.clone()]);
        }
    }

    #[test]
    fn append2_read2_with_corruption() {
        let temp_file = tempfile::NamedTempFile::new().unwrap();
        let entry = Entry {
            timestamp: Some(7),
            humidity: vec![0.95],
            temp: Some(32.2),
            fan: Some(false),
            humidifer: vec![false],
        };
        let entry2 = Entry {
            timestamp: Some(16),
            humidity: vec![0.92],
            temp: Some(31.2),
            fan: Some(true),
            humidifer: vec![false],
        };
        let entry3 = Entry {
            timestamp: Some(17),
            humidity: vec![0.95],
            temp: Some(31.4),
            fan: Some(false),
            humidifer: vec![true],
        };

        {
            use std::io::Write;
            let mut store = Store::open(temp_file.path()).unwrap();
            store.append(&entry).unwrap();
            store.append(&entry2).unwrap();

            let entries = store.lookup_range(0, None);
            assert_eq!(entries, vec![entry.clone(), entry2.clone()]);

            store.contents.write(&[45, 34, 44]).unwrap();
            store.append(&entry3).unwrap();

            let entries = store.lookup_range(0, None);
            assert_eq!(entries, vec![entry.clone(), entry2.clone(), entry3.clone()]);
        }

        {
            let store = Store::open_read_only(temp_file.path()).unwrap();
            let entries = store.lookup_range(0, None);
            assert_eq!(entries, vec![entry.clone(), entry2.clone()]);
        }
    }

    #[test]
    fn append2_with_corruption_write() {
        let temp_file = tempfile::NamedTempFile::new().unwrap();
        let entry = Entry {
            timestamp: Some(7),
            humidity: vec![0.95],
            temp: Some(32.2),
            fan: Some(false),
            humidifer: vec![false],
        };
        let entry2 = Entry {
            timestamp: Some(16),
            humidity: vec![0.92],
            temp: Some(31.2),
            fan: Some(true),
            humidifer: vec![false],
        };
        let entry3 = Entry {
            timestamp: Some(17),
            humidity: vec![0.95],
            temp: Some(31.4),
            fan: Some(false),
            humidifer: vec![true],
        };

        {
            use std::io::Write;
            let mut store = Store::open(temp_file.path()).unwrap();
            store.append(&entry).unwrap();
            store.append(&entry2).unwrap();
            store.contents.write(&[45, 34, 44]).unwrap();
            let entries = store.lookup_range(0, None);
            assert_eq!(entries, vec![entry.clone(), entry2.clone()]);
        }

        {
            let mut store = Store::open(temp_file.path()).unwrap();
            store.append(&entry3).unwrap();
        }

        {
            let store = Store::open_read_only(temp_file.path()).unwrap();
            let entries = store.lookup_range(0, None);
            assert_eq!(entries, vec![entry.clone(), entry2.clone(), entry3.clone()]);
        }
    }

    #[test]
    fn read_different_ranges() {
        let temp_file = tempfile::NamedTempFile::new().unwrap();
        let entry = Entry {
            timestamp: Some(7),
            humidity: vec![0.95],
            temp: Some(32.2),
            fan: Some(false),
            humidifer: vec![false],
        };
        let entry2 = Entry {
            timestamp: Some(16),
            humidity: vec![0.92],
            temp: Some(31.2),
            fan: Some(true),
            humidifer: vec![false],
        };
        let entry3 = Entry {
            timestamp: Some(17),
            humidity: vec![0.95],
            temp: Some(31.4),
            fan: Some(false),
            humidifer: vec![true],
        };

        let mut store = Store::open(temp_file.path()).unwrap();
        store.append(&entry).unwrap();
        store.append(&entry2).unwrap();
        store.append(&entry3).unwrap();
        let entries = store.lookup_range(4, Some(8));
        assert_eq!(entries, vec![entry.clone()]);
        let entries = store.lookup_range(4, Some(16));
        assert_eq!(entries, vec![entry.clone(), entry2.clone()]);
        let entries = store.lookup_range(8, Some(16));
        assert_eq!(entries, vec![entry2.clone()]);
        let entries = store.lookup_range(4, Some(17));
        assert_eq!(entries, vec![entry.clone(), entry2.clone(), entry3.clone()]);
        let entries = store.lookup_range(8, None);
        assert_eq!(entries, vec![entry2.clone(), entry3.clone()]);
    }
}
